@@ -280,7 +280,9 @@ def help(myopts, havecolor=1):
 		print("              With this option, output such as USE=\"dar -bar -foo\" will instead")
 		print("              be displayed as USE=\"-bar dar -foo\"")
 		print()
-		print("       "+green("--ask")+" ("+green("-a")+" short option)")
+		print("       " + green("--ask") + \
+			" [ %s | %s ] (%s short option)" % \
+			(turquoise("y"), turquoise("n"), green("-a")))
 		desc = "Before performing the action, display what will take place (server info for " + \
 			"--sync, --pretend output for merge, and so forth), then ask " + \
 			"whether to proceed with the action or abort.  Using --ask is more " + \
@@ -307,14 +309,24 @@ def help(myopts, havecolor=1):
 		print() 
 		print("       " + green("--autounmask") + " [ %s | %s ]" % \
 			(turquoise("y"), turquoise("n")))
-		desc = "Automatically unmask packages. If any configuration " + \
+		desc = "Automatically unmask packages and generate package.use " + \
+			"settings as necessary to satisfy dependencies. This " + \
+			"option is enabled by default. If any configuration " + \
 			"changes are required, then they will be displayed " + \
 			"after the merge list and emerge will immediately " + \
 			"abort. If the displayed configuration changes are " + \
 			"satisfactory, you should copy and paste them into " + \
-			"the specified configuration file(s). Currently, " + \
-			"this only works for unstable KEYWORDS masks, " + \
-			"LICENSE masks, and package.use settings."
+			"the specified configuration file(s), or enable the " + \
+			"--autounmask-write option. The " + \
+			"EMERGE_DEFAULT_OPTS variable may be used to " + \
+			"disable this option by default in make.conf(5)."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
+		print("       " + green("--autounmask-write") + " [ %s | %s ]" % \
+			(turquoise("y"), turquoise("n")))
+		desc = "If --autounmask is enabled, changes are written " + \
+			"to config files, respecting CONFIG_PROTECT and --ask."
 		for line in wrap(desc, desc_width):
 			print(desc_indent + line)
 		print()
@@ -325,7 +337,7 @@ def help(myopts, havecolor=1):
 		for line in wrap(desc, desc_width):
 			print(desc_indent + line)
 		print()
-		print("        " + green("--binpkg-respect-use") + " [ %s | %s ]" % \
+		print("       " + green("--binpkg-respect-use") + " [ %s | %s ]" % \
 			(turquoise("y"), turquoise("n")))
 		desc = "Tells emerge to ignore binary packages if their use flags" + \
 			" don't match the current configuration. (default: 'n')"
@@ -381,7 +393,10 @@ def help(myopts, havecolor=1):
 			" option will significantly increase the time taken for dependency" + \
 			" calculations. Note that, unlike the --deep option, the" + \
 			" --complete-graph option does not cause any more packages to" + \
-			" be updated than would have otherwise been updated with the option disabled."
+			" be updated than would have otherwise " + \
+			"been updated with the option disabled. " + \
+			"Using --with-bdeps=y together with --complete-graph makes " + \
+			"the graph as complete as possible."
 		for line in wrap(desc, desc_width):
 			print(desc_indent + line)
 		print()
@@ -414,7 +429,10 @@ def help(myopts, havecolor=1):
 			desc = "Account for library link-level dependencies during " + \
 				"--depclean and --prune actions. This " + \
 				"option is enabled by default. In some cases this can " + \
-				"be somewhat time-consuming."
+				"be somewhat time-consuming. This option is ignored " + \
+				"when FEATURES=\"preserve-libs\" is enabled in " + \
+				"make.conf(5), since any libraries that have " + \
+				"consumers will simply be preserved."
 			for line in wrap(desc, desc_width):
 				print(desc_indent + line)
 			print()
@@ -510,6 +528,15 @@ def help(myopts, havecolor=1):
 		for line in wrap(desc, desc_width):
 			print(desc_indent + line)
 		print()
+		print("       " + green("--misspell-suggestions") + " < %s | %s >" % \
+			(turquoise("y"), turquoise("n")))
+		desc = "Enable or disable misspell suggestions. By default, " + \
+			"emerge will show a list of packages with similar names " + \
+			"when a package doesn't exist. The EMERGE_DEFAULT_OPTS " + \
+			"variable may be used to disable this option by default"
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
 		print("       "+green("--newuse")+" ("+green("-N")+" short option)")
 		desc = "Tells emerge to include installed packages where USE " + \
 			"flags have changed since compilation. This option " + \
@@ -538,16 +565,29 @@ def help(myopts, havecolor=1):
 		print("              ebuilds, or deps you specify on the command-line *will* cause")
 		print("              Portage to remerge the package, even if it is already installed.")
 		print("              Note that Portage won't remerge dependencies by default.")
-		desc = "Also note that this option takes " + \
-			"precedence over options such as --newuse, preventing a package " + \
-			"from being reinstalled even though the corresponding USE flag settings " + \
-			"may have changed."
-		for line in wrap(desc, desc_width):
-			print(desc_indent + line)
 		print() 
 		print("       "+green("--nospinner"))
 		print("              Disables the spinner regardless of terminal type.")
 		print()
+		print("       " + green("--usepkg-exclude") + " " + turquoise("ATOMS"))
+		desc = "A space separated list of package names or slot atoms." + \
+			" Emerge will ignore matching binary packages."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
+		print("       " + green("--rebuild-exclude") + " " + turquoise("ATOMS"))
+		desc = "A space separated list of package names or slot atoms." + \
+			" Emerge will not rebuild matching packages due to --rebuild."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
+		print("       " + green("--rebuild-ignore") + " " + turquoise("ATOMS"))
+		desc = "A space separated list of package names or slot atoms." + \
+			" Emerge will not rebuild packages that depend on matching " + \
+			" packages due to --rebuild."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
 		print("       "+green("--oneshot")+" ("+green("-1")+" short option)")
 		print("              Emerge as normal, but don't add packages to the world profile.")
 		print("              This package will only be updated if it is depended upon by")
@@ -578,11 +618,14 @@ def help(myopts, havecolor=1):
 		print("              printed out accompanied by a '+' for enabled and a '-' for")
 		print("              disabled USE flags.")
 		print()
-		print("       "+green("--quiet")+" ("+green("-q")+" short option)")
+		print("       " + green("--quiet") + \
+			" [ %s | %s ] (%s short option)" % \
+			(turquoise("y"), turquoise("n"), green("-q")))
 		print("              Effects vary, but the general outcome is a reduced or condensed")
 		print("              output from portage's displays.")
 		print()
-		print("       "+green("--quiet-build"))
+		print("       " + green("--quiet-build") + \
+			" [ %s | %s ]" % (turquoise("y"), turquoise("n")))
 		desc = "Redirect all build output to logs alone, and do not " + \
 			"display it on stdout."
 		for line in wrap(desc, desc_width):
@@ -596,6 +639,31 @@ def help(myopts, havecolor=1):
 		for line in wrap(desc, desc_width):
 			print(desc_indent + line)
 		print()
+		print("       " + green("--rebuild-if-new-rev") + " [ %s | %s ]" % \
+			(turquoise("y"), turquoise("n")))
+		desc = "Rebuild packages when dependencies that are " + \
+			"used at both build-time and run-time are built, " + \
+			"if the dependency is not already installed with the " + \
+			"same version and revision."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
+		print("       " + green("--rebuild-if-new-ver") + " [ %s | %s ]" % \
+			(turquoise("y"), turquoise("n")))
+		desc = "Rebuild packages when dependencies that are " + \
+			"used at both build-time and run-time are built, " + \
+			"if the dependency is not already installed with the " + \
+			"same version. Revision numbers are ignored."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
+		print("       " + green("--rebuild-if-unbuilt") + " [ %s | %s ]" % \
+			(turquoise("y"), turquoise("n")))
+		desc = "Rebuild packages when dependencies that are " + \
+			"used at both build-time and run-time are built."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
 		print("       " + green("--rebuilt-binaries") + " [ %s | %s ]" % \
 			(turquoise("y"), turquoise("n")))
 		desc = "Replace installed packages with binary packages that have " + \
@@ -622,6 +690,13 @@ def help(myopts, havecolor=1):
 		print("              not trigger reinstallation when flags that the user has not")
 		print("              enabled are added or removed.")
 		print()
+		print("       " + green("--reinstall-atoms") + " " + turquoise("ATOMS"))
+		desc = "A space separated list of package names or slot atoms. " + \
+			"Emerge will treat matching packages as if they are not " + \
+			"installed, and reinstall them if necessary."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
 		print("       "+green("--root=DIR"))
 		desc = "Set the ROOT environment variable " + \
 			"which is documented in the emerge(1) man page."
@@ -652,8 +727,7 @@ def help(myopts, havecolor=1):
 		print()
 		print("       " + green("--selective") + " [ %s | %s ]" % \
 			(turquoise("y"), turquoise("n")))
-		desc = "This is similar to the --noreplace option, except that it " + \
-			"does not take precedence over options such as --newuse. " + \
+		desc = "This identical to the --noreplace option. " + \
 			"Some options, such as --update, imply --selective. " + \
 			"Use --selective=n if you want to forcefully disable " + \
 			"--selective, regardless of options like --update."
@@ -704,6 +778,13 @@ def help(myopts, havecolor=1):
 		for line in wrap(desc, desc_width):
 			print(desc_indent + line)
 		print()
+		print("       " + green("--useoldpkg-atoms") + " " + turquoise("ATOMS"))
+		desc = "A space separated list of package names or slot atoms." + \
+			" Emerge will prefer matching binary packages over newer" + \
+			" unbuilt packages."
+		for line in wrap(desc, desc_width):
+			print(desc_indent + line)
+		print()
 		print("       " + green("--usepkg") + \
 			" [ %s | %s ] (%s short option)" % \
 			(turquoise("y"), turquoise("n"), green("-k")))
