@@ -1,4 +1,4 @@
-# Copyright 2010 Gentoo Foundation
+# Copyright 2010-2011 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
 
 from __future__ import print_function
@@ -128,9 +128,12 @@ class slot_conflict_handler(object):
 
 		self._prepare_conflict_msg_and_check_for_specificity()
 
-		#a list of dicts that hold the needed USE changes to solve all conflicts
+		#a list of dicts that hold the needed USE values to solve all conflicts
 		self.solutions = []
-		
+
+		#a list of dicts that hold the needed USE changes to solve all conflicts
+		self.changes = []
+
 		#configuration = a list of packages with exactly one package from every
 		#single slot conflict
 		config_gen = _configuration_generator(conflict_pkgs)
@@ -152,6 +155,7 @@ class slot_conflict_handler(object):
 
 			if new_solutions:
 				self.solutions.extend(new_solutions)
+
 				if first_config:
 					#If the "all ebuild"-config gives a solution, use it.
 					#Otherwise enumerate all other solutions.
@@ -168,10 +172,63 @@ class slot_conflict_handler(object):
 					writemsg("\nAborting search due to excessive number of configurations.\n", noiselevel=-1)
 				break
 
+		for solution in self.solutions:
+			self._add_change(self._get_change(solution))
+
 
 	def get_conflict(self):
 		return "".join(self.conflict_msg)
-		
+
+	def _is_subset(self, change1, change2):
+		"""
+		Checks if a set of changes 'change1' is a subset of the changes 'change2'.
+		"""
+		#All pkgs of change1 have to be in change2.
+		#For every package in change1, the changes have to be a subset of
+		#the corresponding changes in change2.
+		for pkg in change1:
+			if pkg not in change2:
+				return False
+
+			for flag in change1[pkg]:
+				if flag not in change2[pkg]:
+					return False
+				if change1[pkg][flag] != change2[pkg][flag]:
+					return False
+		return True
+
+	def _add_change(self, new_change):
+		"""
+		Make sure to keep only minimal changes. If "+foo", does the job, discard "+foo -bar".
+		"""
+		changes = self.changes
+		#Make sure there is no other solution that is a subset of the new solution.
+		ignore = False
+		to_be_removed = []
+		for change in changes:
+			if self._is_subset(change, new_change):
+				ignore = True
+				break
+			elif self._is_subset(new_change, change):
+				to_be_removed.append(change)
+
+		if not ignore:
+			#Discard all existing change that are a superset of the new change.
+			for obsolete_change in to_be_removed:
+				changes.remove(obsolete_change)
+			changes.append(new_change)
+
+	def _get_change(self, solution):
+		_pkg_use_enabled = self.depgraph._pkg_use_enabled
+		new_change = {}
+		for pkg in solution:
+			for flag, state in solution[pkg].items():
+				if state == "enabled" and flag not in _pkg_use_enabled(pkg):
+					new_change.setdefault(pkg, {})[flag] = True
+				elif state == "disabled" and flag in _pkg_use_enabled(pkg):
+					new_change.setdefault(pkg, {})[flag] = False
+		return new_change
+
 	def _prepare_conflict_msg_and_check_for_specificity(self):
 		"""
 		Print all slot conflicts in a human readable way.
@@ -231,18 +288,28 @@ class slot_conflict_handler(object):
 								collision_reasons[("version", sub_type)] = atoms
 							elif not atom_set.findAtomForPackage(other_pkg, \
 								modified_use=_pkg_use_enabled(other_pkg)):
-								#Use conditionals not met.
-								violated_atom = atom.violated_conditionals(_pkg_use_enabled(other_pkg), \
-									other_pkg.iuse.is_valid_flag)
-								for flag in violated_atom.use.enabled.union(violated_atom.use.disabled):
-									atoms = collision_reasons.get(("use", flag), set())
-									atoms.add((ppkg, atom, other_pkg))
-									collision_reasons[("use", flag)] = atoms
-								num_all_specific_atoms += 1
+								missing_iuse = other_pkg.iuse.get_missing_iuse(
+									atom.unevaluated_atom.use.required)
+								if missing_iuse:
+									for flag in missing_iuse:
+										atoms = collision_reasons.get(("use", flag), set())
+										atoms.add((ppkg, atom, other_pkg))
+										collision_reasons[("use", flag)] = atoms
+									num_all_specific_atoms += 1
+								else:
+									#Use conditionals not met.
+									violated_atom = atom.violated_conditionals(_pkg_use_enabled(other_pkg), \
+										other_pkg.iuse.is_valid_flag)
+									for flag in violated_atom.use.enabled.union(violated_atom.use.disabled):
+										atoms = collision_reasons.get(("use", flag), set())
+										atoms.add((ppkg, atom, other_pkg))
+										collision_reasons[("use", flag)] = atoms
+									num_all_specific_atoms += 1
 
 					msg.append(" pulled in by\n")
 
 					selected_for_display = set()
+					unconditional_use_deps = set()
 
 					for (type, sub_type), parents in collision_reasons.items():
 						#From each (type, sub_type) pair select at least one atom.
@@ -269,47 +336,66 @@ class slot_conflict_handler(object):
 							#not possible to change them on the parent, which means there
 							#are fewer possible solutions.
 							use = sub_type
-							hard_matches = set()
-							conditional_matches = set()
 							for ppkg, atom, other_pkg in parents:
-								parent_use = None
-								if isinstance(ppkg, Package):
-									parent_use = _pkg_use_enabled(ppkg)
-								violated_atom = atom.unevaluated_atom.violated_conditionals( \
-									_pkg_use_enabled(other_pkg), other_pkg.iuse.is_valid_flag,
-									parent_use=parent_use)
-								if use in violated_atom.use.enabled.union(violated_atom.use.disabled):
-									hard_matches.add((ppkg, atom))
+								missing_iuse = other_pkg.iuse.get_missing_iuse(
+									atom.unevaluated_atom.use.required)
+								if missing_iuse:
+									unconditional_use_deps.add((ppkg, atom))
 								else:
-									conditional_matches.add((ppkg, atom))
-
-							if hard_matches:
-								matches = hard_matches
-							else:
-								matches = conditional_matches
-							
-							if not selected_for_display.intersection(matches):
-								selected_for_display.add(matches.pop())
+									parent_use = None
+									if isinstance(ppkg, Package):
+										parent_use = _pkg_use_enabled(ppkg)
+									violated_atom = atom.unevaluated_atom.violated_conditionals(
+										_pkg_use_enabled(other_pkg),
+										other_pkg.iuse.is_valid_flag,
+										parent_use=parent_use)
+									# It's possible for autounmask to change
+									# parent_use such that the unevaluated form
+									# of the atom now matches, even though the
+									# earlier evaluated form (from before
+									# autounmask changed parent_use) does not.
+									# In this case (see bug #374423), it's
+									# expected that violated_atom.use is None.
+									# Since the atom now matches, we don't want
+									# to display it in the slot conflict
+									# message, so we simply ignore it and rely
+									# on the autounmask display to communicate
+									# the necessary USE change to the user.
+									if violated_atom.use is None:
+										continue
+									if use in violated_atom.use.enabled or \
+										use in violated_atom.use.disabled:
+										unconditional_use_deps.add((ppkg, atom))
+								# When USE flags are removed, it can be
+								# essential to see all broken reverse
+								# dependencies here, so don't omit any.
+								# If the list is long, people can simply
+								# use a pager.
+								selected_for_display.add((ppkg, atom))
 
 					def highlight_violations(atom, version, use=[]):
 						"""Colorize parts of an atom"""
 						atom_str = str(atom)
 						if version:
 							op = atom.operator
-							ver = cpv_getversion(atom.cpv)
+							ver = None
+							if atom.cp != atom.cpv:
+								ver = cpv_getversion(atom.cpv)
 							slot = atom.slot
 
 							if op == "=*":
 								op = "="
 								ver += "*"
 
-							atom_str = atom_str.replace(op, colorize("BAD", op), 1)
-							
-							start = atom_str.rfind(ver)
-							end = start + len(ver)
-							atom_str = atom_str[:start] + \
-								colorize("BAD", ver) + \
-								atom_str[end:]
+							if op is not None:
+								atom_str = atom_str.replace(op, colorize("BAD", op), 1)
+
+							if ver is not None:
+								start = atom_str.rfind(ver)
+								end = start + len(ver)
+								atom_str = atom_str[:start] + \
+									colorize("BAD", ver) + \
+									atom_str[end:]
 							if slot:
 								atom_str = atom_str.replace(":" + slot, colorize("BAD", ":" + slot))
 						
@@ -330,7 +416,14 @@ class slot_conflict_handler(object):
 						
 						return atom_str
 
-					for parent_atom in selected_for_display:
+					# Show unconditional use deps first, since those
+					# are more problematic than the conditional kind.
+					ordered_list = list(unconditional_use_deps)
+					if len(selected_for_display) > len(unconditional_use_deps):
+						for parent_atom in selected_for_display:
+							if parent_atom not in unconditional_use_deps:
+								ordered_list.append(parent_atom)
+					for parent_atom in ordered_list:
 						parent, atom = parent_atom
 						msg.append(2*indent)
 						if isinstance(parent,
@@ -387,7 +480,7 @@ class slot_conflict_handler(object):
 		if self.conflict_is_unspecific and \
 			not ("--newuse" in self.myopts and "--update" in self.myopts):
 			msg += "!!! Enabling --newuse and --update might solve this conflict.\n"
-			msg += "!!! If not, it might at least allow emerge to give a suggestions.\n\n"
+			msg += "!!! If not, it might help emerge to give a more specific suggestion.\n\n"
 			return msg
 
 		solutions = self.solutions
@@ -407,26 +500,26 @@ class slot_conflict_handler(object):
 				msg += "It might be possible to solve these slot collisions\n"
 			msg += "by applying one of the following solutions:\n"
 
-		def print_solution(solution, indent=""):
+		def print_change(change, indent=""):
 			mymsg = ""
-			for pkg in solution:
+			for pkg in change:
 				changes = []
-				for flag, state in solution[pkg].items():
-					if state == "enabled" and flag not in _pkg_use_enabled(pkg):
+				for flag, state in change[pkg].items():
+					if state:
 						changes.append(colorize("red", "+" + flag))
-					elif state == "disabled" and flag in _pkg_use_enabled(pkg):
+					else:
 						changes.append(colorize("blue", "-" + flag))
-				if changes:
-					mymsg += indent + "- " + pkg.cpv + " (Change USE: %s" % " ".join(changes) + ")\n"
+				mymsg += indent + "- " + pkg.cpv + " (Change USE: %s" % " ".join(changes) + ")\n"
 			mymsg += "\n"
 			return mymsg
 
-		if len(solutions) == 1:
-			msg += print_solution(solutions[0], "   ")
+
+		if len(self.changes) == 1:
+			msg += print_change(self.changes[0], "   ")
 		else:
-			for solution in solutions:
+			for change in self.changes:
 				msg += "  Solution: Apply all of:\n"
-				msg += print_solution(solution, "     ")
+				msg += print_change(change, "     ")
 
 		return msg
 
@@ -495,6 +588,15 @@ class slot_conflict_handler(object):
 				else:
 					violated_atom = atom.unevaluated_atom.violated_conditionals(_pkg_use_enabled(pkg), \
 						pkg.iuse.is_valid_flag, parent_use=_pkg_use_enabled(ppkg))
+					if violated_atom.use is None:
+						# It's possible for autounmask to change
+						# parent_use such that the unevaluated form
+						# of the atom now matches, even though the
+						# earlier evaluated form (from before
+						# autounmask changed parent_use) does not.
+						# In this case (see bug #374423), it's
+						# expected that violated_atom.use is None.
+						continue
 
 				if pkg.installed and (violated_atom.use.enabled or violated_atom.use.disabled):
 					#We can't change USE of an installed package (only of an ebuild, but that is already
